
/** @file spi_sam7.c
 * @brief SPI-Library for the AT91SAM7 series of ARM microcontroller.
 * @author Clemens Helfmeier
 * @date 25.06.2006
 *
 * This is the SPI-Library for AT91SAM7 based chips. The current Silicon has
 * a few very bad bugs so, this isn't very efficient for some
 * configurations yet.
 *
 * Documentation of functions in this file only refer to the AT91SAM7
 * specific part of the functions and their usage. For more common
 * documentation see spi.c
 * 
 * ATTENTION:
 * This library works with the usual devices. It supports CSAAT as
 * long as the next datum ist put into the library before the last
 * one ends. The library has a fifo of one datum length. Thus the
 * usual procedure would be to put exactly two characters into the
 * lib before reading the answer of the first one.
 *
 * spi_tmit won't wait until the character is sent. it just checks
 * for the sending buffer to be empty.
 * spi_rmit will wait until one character is in the reciever
 * buffer, note that that one is not nesseccaritly the last one
 * recieved.
 * spi_xmit will always send exactly one datum and return exactly
 * that datum that was the answer for the datum sent before.
 *
 * You should keep track of the number of bytes in the reciever
 * fifo and not let it exceed 2 since this will result in an
 * overflow and in loosing synchronisation.
 * Always read the answers if you are interesed in the results of
 * bytes later transferred.
 *
 * This is due to the not working CSAAT of the controller.
 *
 * The block transfer functions spi_bxmit, spi_btmit can be used in
 * conjunction with all the other functions that only transfer a single
 * word. There is no need to equalize the read and written number of
 * words before calling the block functions. They will take care to
 * exactly read the bytes they also send.
 *
 * spi_bxmit is not yet fully beta-tested. So be aware that there could
 * be a bug in the lib.
 *
 * @see spi.c
 */
 
#include "AT91SAM7S64.h"
#define __inline extern inline
#include "lib_AT91SAM7S64.h"

#include "spi.h"
#include "config_spi.h"

/// Debugging support
#include "debug.h"
#define DBG_LEVEL	DBG_LEVEL_ERROR
#include "debug.h"

/** Use the Peripheral Device Controller (PDC). */
#define USE_DMA	1
/** Enable the task_switch() function in spi_wait(). */
//#define ENABLE_TASK_CHANGE	1
#ifdef ENABLE_TASK_CHANGE
#include "task.h"
#endif

/// @private
/// active chip number (used in block functions)
static unsigned char active_cs = -1;
const unsigned int cs_modes[ 4 ] = { SPI_CS_0, SPI_CS_1, SPI_CS_2, SPI_CS_3 };

/**
 * Initialize all the AT91SAM7 stuff here, deselect all slaves.
 */
void spi_init( void ) {
	/* start the clock for SPI */
	AT91F_SPI_CfgPMC();
	/* Reset SPI */
	AT91F_SPI_Reset( AT91C_BASE_SPI );
	/* initialize PIO to drive IO signals */
	AT91F_PIO_CfgPeriph( AT91C_BASE_PIOA, AT91C_PA13_MOSI | AT91C_PA12_MISO | AT91C_PA14_SPCK, 0 );
	/* initialize PIO to drive chip select signals */
	AT91F_PIO_CfgPeriph( AT91C_BASE_PIOA, SPI_CSLINES_A, SPI_CSLINES_B );
	/* for software chip select, put all the lines up */
	AT91F_PIO_SetOutput( AT91C_BASE_PIOA, SPI_CS_0 | SPI_CS_1 | SPI_CS_2 | SPI_CS_3 );
	AT91F_PIO_CfgOutput( AT91C_BASE_PIOA, SPI_CS_0 | SPI_CS_1 | SPI_CS_2 | SPI_CS_3 );
	/* stop the PDC */
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTDIS | AT91C_PDC_RXTEN;
	AT91C_BASE_PDC_SPI->PDC_RCR == 0;
	AT91C_BASE_PDC_SPI->PDC_TCR == 0;
	AT91C_BASE_PDC_SPI->PDC_RNCR == 0;
	AT91C_BASE_PDC_SPI->PDC_TNCR == 0;
	/* enable SPI */
	AT91F_SPI_Enable( AT91C_BASE_SPI );
	/* configure Mode Register */
	AT91C_BASE_SPI->SPI_MR =
			AT91C_SPI_MSTR |	/* we use MASTER only mode */
			AT91C_SPI_PS_FIXED | 	/* we only use Fixed Peripheral Select */
			AT91C_SPI_MODFDIS | 	/* no Fault protection since there is no second master */
			AT91C_SPI_PCS | 	/* no peripheral selected */
			((10<<24) & AT91C_SPI_DLYBCS);	/* 10 Clock Cycles betwenn two different chip selects */
	/* program all chip select registers correctly */
	AT91C_BASE_SPI->SPI_CSR[0] = SPI_CSR_0;
	AT91C_BASE_SPI->SPI_CSR[1] = SPI_CSR_1;
	AT91C_BASE_SPI->SPI_CSR[2] = SPI_CSR_2;
	AT91C_BASE_SPI->SPI_CSR[3] = SPI_CSR_3;
	active_cs = -1;
}

void spi_start( void* dev ) {
	/* wait for any previous transfer to be finished */
	if ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_SPIENS) != 0) {
		/* wait until all transfers are finished */
		while (((AT91C_BASE_SPI->SPI_SR & (AT91C_SPI_TXEMPTY)) != (AT91C_SPI_TXEMPTY)));
		/* remove any left bytes inside the RDR */
		while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF) != 0) AT91C_BASE_SPI->SPI_RDR;
	}
	/* disable PDC */
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS;
	/* dev isn't a pointer here, it is just the number of the chip select line to be used */
	active_cs = (int)dev;
	AT91C_BASE_SPI->SPI_MR = (AT91C_BASE_SPI->SPI_MR & (~AT91C_SPI_PCS)) | ((~((1<<(active_cs)<<16))) & AT91C_SPI_PCS);
	/* make shure the SPI is enabled */
	AT91F_SPI_Enable( AT91C_BASE_SPI );
	/* pull cs line low if active */
	if (cs_modes[ active_cs ] != 0)
		AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, cs_modes[ active_cs ] );
}

void spi_end( void ) {
	/* now set the LASTXFER value */
	AT91C_BASE_SPI->SPI_CR = AT91C_SPI_LASTXFER;
	/* if the cs line is controlled in software, wait for transfers finished and deselect the device here. */
	if (cs_modes[ active_cs ] != 0) {
#ifdef USE_DMA
		/* wait for any DMA transfer to be finished */
		while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXBUFE) == 0);
#endif
		while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXEMPTY) == 0);
		AT91F_PIO_SetOutput( AT91C_BASE_PIOA, cs_modes[ active_cs ] );
	}
	active_cs = -1;
}

unsigned short spi_xmit( unsigned short send ) {
	unsigned short retval;
#ifdef USE_DMA
	/* wait for any DMA transfer to be finished */
	while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXBUFE) == 0);
#endif
	/* wait until send is empty */
	while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TDRE) == 0);
	if ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF) != 0) retval = AT91C_BASE_SPI->SPI_RDR & 0xFFFF;
	/* send the character */
	AT91C_BASE_SPI->SPI_TDR = (send & 0xFFFF);
	/* wait until (our) data is recieved */
	do {
		if ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF) != 0) {
			retval = AT91C_BASE_SPI->SPI_RDR & 0xFFFF;
		}
	} while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXEMPTY) == 0);
	/* return the last read data */
	return retval;
}

void spi_tmit( unsigned short send ) {
#ifdef USE_DMA
	/* wait for any DMA transfer to be finished */
	while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXBUFE) == 0);
#endif
	while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TDRE) == 0);
	/* send the character */
	AT91C_BASE_SPI->SPI_TDR = (send & 0xFFFF);
}

unsigned short spi_rmit( void ) {
	/* wait until data is recieved */
	while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF) == 0);
	/* return the read data */
	return AT91C_BASE_SPI->SPI_RDR & 0xFFFF;
}

/** @test
 * This function must still be tested. Currently working is the following:
 * PDC idle,
 * normal PDC (read and write) transfer before.
 */
void spi_bxmit( void *data, unsigned short len ) {
#ifdef USE_DMA
	if ((AT91C_BASE_PDC_SPI->PDC_PTSR & AT91C_PDC_TXTEN) != 0) while ((AT91C_BASE_PDC_SPI->PDC_TNCR) != 0);
	/* When using the PDC, we only need to re-synchronize recieving and transmitting
	data if the previous transfer wasn't a PDC transfer. If there is still a PDC transfer
	running (say PDC_RCR != 0), the transfer is synchron. This is based upon the assumption
	that this library won't do any PDC transfer that recieves more or less bytes than it transmits.
	In case of a previous Transmit-Only transfer, PDC_TCR will be higher than zero and PDC_RCR will be zero.
	We can simply fill in PDC_TNCR and wait for PDC_TNCR to be zero. Then we read the RHR once (clearing
	all previously recieved data) and put the RCR up.
	*/
	if ((AT91C_BASE_PDC_SPI->PDC_PTSR & AT91C_PDC_TXTEN) == 0)
		AT91C_BASE_PDC_SPI->PDC_TCR == 0;
	if ((AT91C_BASE_PDC_SPI->PDC_PTSR & AT91C_PDC_RXTEN) == 0)
		AT91C_BASE_PDC_SPI->PDC_RCR == 0;
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS;	// disable transfer
	if (((AT91C_BASE_PDC_SPI->PDC_TCR) == 0) && ((AT91C_BASE_PDC_SPI->PDC_RCR) == 0)) {
		/* the PDC for this channel is idle. */
		/* in case there is still a previous transfer with the single datum functions active,
		start this PDC transfer in a bit different way. */
		/* copy the buffer and number of transfers to the reciever, recieving won't start
		since the reciever is disabled!
		*/
		AT91C_BASE_PDC_SPI->PDC_RPR = (int)data;
		AT91C_BASE_PDC_SPI->PDC_RCR = len;
		/* write one datum */
		if ((AT91C_BASE_SPI->SPI_CSR[active_cs] & AT91C_SPI_BITS) > (AT91C_SPI_BITS_8<<4)) {
			/* write 16 bits */
			spi_tmit( *((unsigned short*)data) );
			data += 2;
		} else {
			/* write 8 bits */
			spi_tmit( *((unsigned char*)data) );
			data += 1;
		}
		len--;
		/* configure TX channel */
		AT91C_BASE_PDC_SPI->PDC_TPR = (int)data;
		AT91C_BASE_PDC_SPI->PDC_TCR = len;
		/* wait until the datum is transferred to the shift register (and any previous recieved data
		is put into the RDR... */
		while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TDRE) == 0);
		/* read any previous RDR value */
		while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF) != 0) AT91C_BASE_SPI->SPI_RDR;
	} else {
		/* PDC is not idle */
		if (((AT91C_BASE_PDC_SPI->PDC_TCR) == 0) && ((AT91C_BASE_PDC_SPI->PDC_RCR) != 0)) {
			/* still recieving the last byte from a previous normal PDC transfer */
			/* fill the transmit registers */
			AT91C_BASE_PDC_SPI->PDC_TPR = (int)data;
			AT91C_BASE_PDC_SPI->PDC_TCR = len;
			/* wait for the previous recieve data to be finished */
			AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTEN;
			AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_RXTEN;
			while (AT91C_BASE_PDC_SPI->PDC_RCR != 0);
			AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTDIS;
			AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_RXTDIS;
			/* fill the recieve register */
			AT91C_BASE_PDC_SPI->PDC_RPR = (int)data;
			AT91C_BASE_PDC_SPI->PDC_RCR = len;
		} else if (((AT91C_BASE_PDC_SPI->PDC_TCR) != 0) && ((AT91C_BASE_PDC_SPI->PDC_RCR) == 0)) {
			/* previous transfer was transmit-only and is still running */
			/* fill next transmit registers */
			AT91C_BASE_PDC_SPI->PDC_TNPR = (int)data;
			AT91C_BASE_PDC_SPI->PDC_TNCR = len;
			/* wait until the TNCR is zero (the registers are transferred to the normal PDC registers, the previous
			transfer is finished. */
			AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTEN;
			while (AT91C_BASE_PDC_SPI->PDC_TNCR != 0);
			AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTDIS;
			AT91C_BASE_SPI->SPI_RDR;	// synchronize by reading the RDR
			/* fill in the Recieve registers (normal register!) */
			AT91C_BASE_PDC_SPI->PDC_RCR = len;
			AT91C_BASE_PDC_SPI->PDC_RPR = (int)data;
		} else if (((AT91C_BASE_PDC_SPI->PDC_TCR) != 0) && ((AT91C_BASE_PDC_SPI->PDC_RCR) != 0)) {
			/* previous transfer was normal, only put to next PDC registers */
			AT91C_BASE_PDC_SPI->PDC_RNPR = (int)data;
			AT91C_BASE_PDC_SPI->PDC_TNPR = (int)data;
			AT91C_BASE_PDC_SPI->PDC_RNCR = len;
			AT91C_BASE_PDC_SPI->PDC_TNCR = len;
		}
	}
	/* start transfer by enabling */
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTEN;
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_RXTEN;
#else
	spi_tmit( *((unsigned char*)data) );
	len--;
	while (len-- > 0) {
		spi_tmit( *((unsigned char*)data+1) );
		*(unsigned char*)data++ = spi_rmit();
	}
	*(unsigned char*)data++ = spi_rmit();
#endif
}

void spi_btmit( void *data, unsigned short len ) {
#ifdef USE_DMA
	if ((AT91C_BASE_PDC_SPI->PDC_PTSR & AT91C_PDC_TXTEN) != 0) while ((AT91C_BASE_PDC_SPI->PDC_TNCR) != 0);
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS;
	if (((AT91C_BASE_PDC_SPI->PDC_TCR) == 0) && ((AT91C_BASE_PDC_SPI->PDC_RCR) == 0)) {
		/* write to TR */
		AT91C_BASE_PDC_SPI->PDC_TPR = (int)data;
		AT91C_BASE_PDC_SPI->PDC_RCR = 0;
		AT91C_BASE_PDC_SPI->PDC_TCR = len;
	} else {
		/* write to NTR */
		AT91C_BASE_PDC_SPI->PDC_TNPR = (int)data;
		AT91C_BASE_PDC_SPI->PDC_RNCR = 0;
		AT91C_BASE_PDC_SPI->PDC_TNCR = len;
	}
	/* enable (and start) the transfer */
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTEN;
#else
	spi_tmit( *((unsigned char*)data) );
	len--;
	while (len-- > 0) {
		spi_tmit( *((unsigned char*)data+1) );
		spi_rmit();
		data+=1;
	}
	spi_rmit();
#endif
}

void spi_wait( void ) {
#ifdef USE_DMA
#ifdef ENABLE_TASK_CHANGE
	task_wait_mem_notequal( &(AT91C_BASE_SPI->SPI_SR), (AT91C_SPI_RXBUFF | AT91C_SPI_TXBUFE), 0 );
#else
	while ((AT91C_BASE_SPI->SPI_SR & (AT91C_SPI_RXBUFF | AT91C_SPI_TXBUFE)) == 0);
#endif
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_RXTDIS;
	AT91C_BASE_PDC_SPI->PDC_PTCR = AT91C_PDC_TXTDIS;
#else
	while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXEMPTY) == 0);
#endif
}
